#!/usr/bin/env python 
# -*- coding: utf-8 -*-

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
define the plug-in architecture

Use :class:`spec2nexus.plugin.ControlLineHandler` as a metaclass
to create a plugin handler class for each SPEC control line.  
In each such class, it is necessary to:

* define a string value for the ``key`` (class attribute)
* override the definition of :meth:`process`

It is optional to:

* define :meth:`postprocess`
* define :meth:`writer`
* define :meth:`match_key`

.. rubric:: Classes

.. autosummary::

  ~ControlLineHandler
  ~PluginManager

.. rubric:: Exceptions

.. autosummary::

  ~DuplicateControlLineKey
  ~DuplicateControlLinePlugin
  ~DuplicatePlugin
  ~PluginBadKeyError
  ~PluginDuplicateKeyError
  ~PluginKeyNotDefined
  ~PluginProcessMethodNotDefined

"""


from collections import OrderedDict
import logging
import re


logger = logging.getLogger(__name__)


class PluginException(Exception):
    """parent exception for this module"""

class DuplicateControlLinePlugin(PluginException): 
    """This control line handler has been used more than once."""

class DuplicateControlLineKey(PluginException): 
    """This control line key regular expression has been used more than once."""

class DuplicatePlugin(PluginException): 
    """This plugin file name has been used more than once."""

class PluginKeyNotDefined(PluginException): 
    """Must define 'key' in class declaration."""

class PluginProcessMethodNotDefined(PluginException): 
    """Must define 'process()' method in class declaration."""

class PluginDuplicateKeyError(PluginException): 
    """This plugin key has been used before."""

class PluginBadKeyError(PluginException): 
    """The plugin 'key' value is not acceptable."""


plugin_manager = None


def get_plugin_manager():
    """
    get the instance of the plugin_manager (a singleton)
    
    Create instance of PluginManager() if necessary.
    Also, 
    """
    global plugin_manager
    if plugin_manager is None:
        plugin_manager = PluginManager()
        if len(plugin_manager.registry) == 0:
            # Now is the time to load all plugins
            plugin_manager.load_plugins()
    return plugin_manager


class AutoRegister(type):

    """
    plugin to handle a single control line in a SPEC data file
    
    This class is a metaclass to auto-register plugins to handle
    various parts of a SPEC data file.  
    See :mod:`~spec2nexus.plugins.spec_common` for many examples.

    :param str key: regular expression to match a control line key, up to the first space
    :returns: None
    """
    
    key = None

    def __init__(cls, *args):       # args: name, bases, dict
        # logger.debug(" "*4 + "."*10)
        # logger.debug(f"__init__: cls={cls}")
        # logger.debug(f"__init__: name={name}")
        # logger.debug(f"__init__: bases={bases}")
        # logger.debug(f"__init__: dict={dict}")
        manager = get_plugin_manager()
        manager.register_control_line_handler(cls)

    def __new__(metaname, classname, baseclasses, attrs):
        # logger.debug(" "*4 + "."*10)
        # logger.debug(f'__new__: metaname={metaname}')
        # logger.debug(f'__new__: classname={classname}')
        # logger.debug(f'__new__: baseclasses={baseclasses}')
        # logger.debug(f'__new__: attrs={attrs}')
        return type.__new__(metaname, classname, baseclasses, attrs)

    def __str__(self):
        return str(self.__name__)


class ControlLineHandler(object):
    """
    base class for SPEC data file control line handler plugins
    
    define one ControlLineHandler class for each different type of control line

    :param str key: regular expression to match a control line key, up to the first space
    :param [str] scan_attributes_defined: list of scan attributes defined in this class
    :returns: None

    EXAMPLE of ``match_key`` method:
    
    Declaration of the ``match_key`` method is optional in a subclass.
    This is used to test a given line from a SPEC data file against the
    ``key`` of each ``ControlLineHandler``.
    
    If this method is defined in the subclass, it will be called
    instead of :meth:`~spec2nexus.plugin.PluginManager.match_key()`.
    This is the example used by 
    :class:`~spec2nexus.plugins.spec_common.SPEC_DataLine`::

        def match_key(self, text):
            try:
                float( text.strip().split()[0] )
                return True
            except ValueError:
                return False
    """
    key = None
    scan_attributes_defined = []
    
    def process(self, text, spec_file_obj, *args, **kws):
        """*required:* handle this line from a SPEC data file"""
        raise NotImplementedError("must override in subclass")

    def postprocess(self, header, *args, **kws):
        """*optional:* additional processing deferred until *after* data file has been read"""
        raise NotImplementedError("must override in subclass")

    def writer(self, h5parent, writer, scan, nxclass=None, *args, **kws):
        """*optional:* Describe how to store this data in an HDF5 NeXus file"""
        raise NotImplementedError("must override in subclass")


class PluginManager(object):

    """
    Manage the set of SPEC data file control line plugins

    .. rubric:: Class Methods
    
    .. autosummary::
    
      ~get
      ~getKey
      ~get_registry
      ~get_registry_table
      ~load_plugins
      ~match_key
      ~process
      ~register_control_line_handler
  
    """
 
    def __init__(self):
        global plugin_manager
        if plugin_manager is None:
            plugin_manager = self
        self.registry = OrderedDict() # dictionary of known ControlLineHandler subclasses
        self.lazy_attributes = []

    def load_plugins(self):
        """
        load all spec2nexus plugin modules
        
        called from :func:`spec2nexus.plugin.get_plugin_manager()`
        """
        from . import spec
        from . import plugins   # issue #166: plugins are loaded here, NOT any earlier!
        
        table = self.get_registry_table()
        logger.debug(str(table))
    
        return self
    
    def getKey(self, spec_data_file_line):
        """
        Find the key that matches this line in a SPEC data file.  Return None if not found.
        
        :param str spec_data_file_line: one line from a SPEC data file
        """
        pos = spec_data_file_line.find(' ')
        if pos < 0:
            return None
        text = spec_data_file_line[:pos]
        
        # try to locate the key directly
        if text in self.registry:
            return text
        
        # brute force search and match using regular expressions
        return self.match_key(text)

    def match_key(self, text):
        """
        test if any handler's key matches text
        
        :param str text: first word on the line, 
            up to but not including the first whitespace
        :returns: key or None
        
        Applies a regular expression match using each handler's
        ``key`` as the regular expression to match with ``text``.
        """
        def _match_(text, handler):
            try:
                if handler().match_key(text):
                    return handler.key
            except AttributeError:
                # ensure that #X and #XPCS do not both match #X
                full_pattern = '^' + handler.key + '$'
                t = re.match(full_pattern, text)
                # test regexp match to avoid false positives
                # ensures that beginning and end are different positions
                return t and t.regs[0][1] != t.regs[0][0]

        for key, handler in self.registry.items():
            if _match_(text, handler):
                return key
        
        return None

    def get(self, key):
        """return the handler identified by key or None"""
        return self.registry.get(key)
    
    def process(self, key, *args, **kw):
        """pick the control line handler by key and call its process() method"""
        handler = self.get(key)
        if handler is not None:
            handler().process(*args, **kw)

    def get_registry(self):
        return self.registry

    def get_registry_table(self, print_it=False):
        """return a table of all the known plugins"""
        import pyRestTable
        tbl = pyRestTable.Table()
        tbl.addLabel("control line")
        tbl.addLabel("handler class")
        for k, v in self.registry.items():
            tbl.addRow((k, v))
        if print_it:
            print("Plugin registry")
            print(tbl)
        return tbl

    def register_control_line_handler(self, handler):
        """
        auto-registry of all AutoRegister plugins
        
        Called from AutoRegister.__init__
        """
        obj = handler()
    
        if not hasattr(obj, "key") or obj.key is None:
            emsg = "'key' not defined: " + obj.__class__.__name__
            raise PluginKeyNotDefined(emsg)
    
        key = obj.key
    
        if key in self.registry:
            emsg = "duplicate key=%s: %s" % (key, obj.__class__)
            previous = self.registry[key]()
            emsg += ", previously defined: " + previous.__class__.__name__
            raise PluginDuplicateKeyError(emsg)
        
        if len(key.strip().split()) != 1:
            emsg = "badly-formed 'key': received '%d'" % key
            raise PluginBadKeyError(emsg)
    
        if not hasattr(obj, "process") :
            emsg = "'process()' method not defined:" + obj.__class__.__name__
            raise PluginProcessMethodNotDefined(emsg)
    
        for att in obj.scan_attributes_defined:
            if att not in self.lazy_attributes:
                self.lazy_attributes.append(att)
    
        self.registry[key] = handler



#!/usr/bin/env python

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
read a list of SPEC data files (or directories) and plot images of all scans

.. autosummary::

    ~DirectoryNotFoundError
    ~PlotSpecFileScans
    ~Cache_File_Mtime
    ~datePath
    ~getSpecFileDate
    ~needToMakePlot
    ~timestamp
    ~buildIndexHtml
    ~logger

.. rubric:: RESULT

The images are stored in files within a directory structure
that is organized chronologically,
such as: ``yyyy/mm/spec_file/s1.svg``.
The root of the directory is either specified by the 
command line ``-d`` option or  defaults to the current 
working directory.  The ``yyyy/mm`` (year and month) are 
taken from the ``#D`` line of the SPEC data file.
The ``spec_file`` is the file name with file extension 
and directory name removed.  The image file names are
derived from the scan numbers.

.. rubric:: Linux CRON task

This script could be called from a *cron* entry, such as::

    # every five minutes (generates no output from outer script)
    0-59/5 * * * *  /some/directory/specplot_gallery.py -d /web/page/dir /spec/data/file/dir

If this script is called too frequently and the list of plots to be generated
is large enough, it is possible for more than one process to be running.
In one extreme case, many processes were found running due to problems with the data files.
To identify and stop all processes of this program::

    kill -9 `ps -ef | grep python | awk '/specplot_gallery.py/ {print $2}' -`

"""


import datetime
import json
import logging
import os
import shutil
import sys

from . import spec
from . import specplot

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError     # py27 compatibility


MTIME_CACHE_FILE = 'mtime_cache.json'
PLOT_TYPE = ".svg"
HTML_INDEX_FILE = 'index.html'
DOC_URL = 'http://spec2nexus.readthedocs.io/en/latest/specplot_gallery.html'


class DirectoryNotFoundError(ValueError): 
    """Exception: The requested directory does not exist"""

class PathIsNotDirectoryError(ValueError): 
    """Exception: The path is not a directory"""


class PlotSpecFileScans(object):
    
    """
    read a SPEC data file and plot thumbnail images of all its scans
    
    :param [str] filelist: list of SPEC data files to be checked
    :param str plotDir: name of base directory to store output image thumbnails

    .. autosummary::
    
        ~specFileUpdated
        ~plot_all_scans
        ~getPlotDir
        ~getBaseDir
        ~href_format

    """

    def __init__(self, filelist, plotDir = None, reverse_chronological = False):
        self.filelist = filelist
        self.plotDir = plotDir or os.getcwd()
        self.reversed = reverse_chronological

        for specFile in filelist:
            self.plot_all_scans(specFile)
    
    def specFileUpdated(self, specFile):
        """
        report if specFile has been updated
        
        Return mtime cache entry (or `None` if not updated)
        """
        result = None
        if os.path.exists(specFile):
            mtime_cache = Cache_File_Mtime(self.plotDir)
            if mtime_cache.was_file_updated(specFile):
                result = mtime_cache.get(specFile)
        return result
    
    def plot_all_scans(self, specFile):
        """
        plot all the recognized scans from the file named ``specFile``
        """
        if not spec.is_spec_file(specFile):
            raise spec.NotASpecDataFile(specFile)

        try:
            logger("SPEC data file: %s" % specFile)
            sd = specplot.openSpecFile(specFile)
        except FileNotFoundError:
            return    # could not open file, be silent about it
        if len(sd.headers) == 0:    # no scan header found, again, silence
            return

        plot_path = self.getPlotDir(specFile)
        logger("  plot directory: %s" % plot_path)
        if not os.path.exists(plot_path):
            os.makedirs(plot_path)
            logger('creating directory: ' + plot_path)
        
        # list the plots we expect
        scans = {}
        for scan_n in sd.getScanNumbers():
            spec_scan = sd.getScan(scan_n)
            # make certain that plot files will sort lexically:  S1 --> s00001
            base = 's%05s' + PLOT_TYPE
            basePlotFile = (base % str(spec_scan.scanNum)).replace(' ', '0')
            altText = '#' + str(spec_scan.scanNum) + ': ' + spec_scan.scanCmd
            href = self.href_format(basePlotFile, altText)
            full = os.path.join(plot_path, basePlotFile)
            scans[scan_n] = dict(
                make=True,
                base=(base % str(spec_scan.scanNum)).replace(' ', '0'), 
                full=full,
                href=href,
                exists=os.path.exists(full),
                spec_scan=spec_scan,
                )
        
        # delete any existing plots that must be remade
        last_cache = Cache_File_Mtime(self.plotDir).get(specFile)
        cache = self.specFileUpdated(specFile)
        if cache is None:
            return
        plot_list = [
            k
            for k in sorted(os.listdir(plot_path))
            if k.endswith(PLOT_TYPE)
            ]
        if len(plot_list) > 0 and cache["size"] > 0 and cache["size"] != last_cache["size"]:
            if cache["size"] > last_cache["size"]:
                # Was last scan updated with more data?  Look at file.
                with open(specFile, "r") as fp:
                    # skip the part already considered
                    fp.read(last_cache["size"])
                    # look at the addition
                    buf = fp.read()
 
                # only delete last plot if addition not start with #S
                if not buf.lstrip().startswith("#S "):
                    k = plot_list.pop()
                    os.remove(os.path.join(plot_path, k))
            else:
                # remake all the plots
                for k in plot_list:     # TODO: needs unit test
                    os.remove(os.path.join(plot_path, k))
                plot_list = []

        # make plots as needed
        remake_index_file = False
        problem_scans = []
        for scan in scans.values():
            scan["make"] = scan["base"] not in plot_list
            if scan["make"]:
                try:
                    logger('  creating SPEC data scan image: ' + scan["base"])
                    selector = specplot.Selector()
                    image_maker = selector.auto(scan["spec_scan"])
                    plotter = image_maker()
                    plotter.plot_scan(scan["spec_scan"], scan["full"])
                    remake_index_file = True
                except Exception as _exc_obj:
                    msg = "<b>%s</b>" % type(_exc_obj).__name__
                    msg += ": <tt>#S %s</tt>" % str(scan["spec_scan"].scanNum)
                    #msg += " (%s)" % specFile
                    problem_scans.append(msg)

        # (re)make the index.html file as needed
        htmlFile = os.path.join(plot_path, HTML_INDEX_FILE)
        if remake_index_file or not os.path.exists(htmlFile):
            logger('  creating/updating index.html file')

            # list the plottable files
            def sorter(d):
                return d["spec_scan"].epoch
    
            plot_list = [
                scan["href"]
                for scan in sorted(
                    scans.values(), 
                    key=sorter, 
                    reverse=self.reversed)
                if os.path.exists(scan["full"])
                ]

            html = buildIndexHtml(specFile, plot_list, problem_scans)
            f = open(htmlFile, "w")
            f.write(html)
            f.close()
            
        # touch to update the mtime on the plot_path
        os.utime(plot_path, None)
    
    def getPlotDir(self, specFile):
        """
        return the plot directory based on the specFile
        
        :param str specFile: name of SPEC data file (relative or absolute)
        """
        data_file_root_name = os.path.splitext(os.path.split(specFile)[1])[0]
        date_str = getSpecFileDate(specFile)
        if date_str is None:
            return
        return self.getBaseDir(data_file_root_name, date_str)

    def getBaseDir(self, basename, date):
        """
        find the path based on the date in the spec file
        """
        return os.path.join(self.plotDir, datePath(date), basename)
    
    def href_format(self, basePlotFile, altText):
        href = '<a href="' + basePlotFile + '">'
        href += '<img src="' + basePlotFile + '"'
        href += ' width="150" height="75"'
        href += ' alt="' + altText + '"/>'
        href += '</a>'
        return href


class Cache_File_Mtime(object):
    
    """
    maintain a list of all known data file modification times
    
    :param str base_dir: name of base directory to store output image thumbnails
    
    This list will allow the code to avoid unnecessary work
    reparsing and plotting of unchanged SPEC data files.
    """

    def __init__(self, base_dir):
        if not os.path.exists(base_dir):
            msg = 'directory does not exist: ' + base_dir
            raise DirectoryNotFoundError(msg)
        self.base_dir = os.path.abspath(base_dir)
        self.cache_file = os.path.join(self.base_dir, MTIME_CACHE_FILE)
        self.cache = self.read()

    def read(self):
        """read the cache from storage"""
        cache = {}
        if os.path.exists(self.cache_file):
            with open(self.cache_file, 'r') as fp:
                cache = json.load(fp)
        return cache

    def write(self):
        """write the cache to storage"""
        with open(self.cache_file, 'w') as fp:
            json.dump(self.cache, fp, indent=4)
    
    def get(self, fname, default=dict(mtime=0, size=0)):
        """
        get the mtime cache entry for data file ``fname``
         
        :param str fname: file name, already known to exist
        :return: time (float) cached value of when fname was 
            last modified or None if not known
        """
        return self.cache.get(fname, default)
        
    def was_file_updated(self, fname):
        """
        compare the mtime between disk and cache
        
        :param str fname: file name, already known to exist
        :return bool: True if file is newer than the cache (or new to the cache)
        """
        assert os.path.exists(fname)
        updated = False
        cache = self.get(fname)
        
        file_mtime = os.path.getmtime(fname)
        file_size = os.path.getsize(fname)

        if file_mtime != cache["mtime"] and file_size != cache["size"]:
            updated = True

        if updated:
            logger('SPEC data file updated: ' + fname)
            self.cache[fname] = dict(mtime=file_mtime, size=os.path.getsize(fname))
            self.write()
        return updated

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def datePath(date):
    """
    convert the date into a path: yyyy/mm
    
    :param str date: text date from SPEC file #D line: 'Thu Jun 19 12:21:55 2014'
    """
    import time
    dateStr = time.strptime(date, "%a %b %d %H:%M:%S %Y")
    yyyy = "%04d" % dateStr.tm_year
    mm = "%02d" % dateStr.tm_mon
    return os.path.join(yyyy, mm)


def getSpecFileDate(specFile):
    """
    return the #D date of the SPEC data file or None
    
    :param str specFile: name of SPEC data file (relative or absolute)
    """
    # Get that without parsing the data file, it's on the 3rd line of the file.
    #     #F 06_19_Tony.dat
    #     #E 1403198515
    #     #D Thu Jun 19 12:21:55 2014
    if not os.path.exists(specFile):
        return None

    # read the first lines of the file and validate for SPEC data file format
    f = open(specFile, 'r')
    line = f.readline()
    if not line.startswith('#F '): return None
    line = f.readline()
    if not line.startswith('#E '): return None
    line = f.readline()
    if not line.startswith('#D '): return None
    f.close()

    return line[2:].strip()  # 'Thu Jun 19 12:21:55 2014'


def needToMakePlot(fullPlotFile, mtime_specFile):
    """
    Determine if a plot needs to be (re)made.  Use mtime as the basis.

    :return bool: ``True`` if plot should be made again
    """
    remake_plot = True
    if os.path.exists(fullPlotFile):
        mtime_plotFile = os.path.getmtime(fullPlotFile)
        if mtime_plotFile > mtime_specFile:
            # plot was made after the data file was updated
            remake_plot = False     # don't remake the plot
    return remake_plot


def timestamp():
    """
    current time as yyyy-mm-dd hh:mm:ss
    
    :return str: 
    """
    ts = str(datetime.datetime.now())
    ts = ' '.join(ts.split('T'))    # convert to modified ISO8601 format
    ts = ts.split('.')[0]           # strip off the fractional seconds
    return ts


def buildIndexHtml(specFile, plotted_scans, problem_scans):
    """
    build index.html content
    
    :param str specFile: name of SPEC data file (relative or absolute)
    :param [str] plotList: list of HTML `<a>` elements, one for each plot image 
    """
    baseSpecFile = os.path.basename(specFile)
    comment = "\n"
    comment += "   written by: %s\n" % sys.argv[0]
    comment += "   date: %s\n"       % timestamp()
    comment += "\n"

    href = "<a href='%s'>%s</a>" % (baseSpecFile, specFile)
    html  = "<html>\n"
    html += "  <head>\n"
    html += "    <title>SPEC scans from %s</title>\n" % os.path.split(specFile)[-1]
    html += "    <!-- %s -->\n"                       % comment
    html += "  </head>\n"
    html += "  <body>\n"
    html += "    <h1>SPEC scans from: %s</h1>\n"      % os.path.split(specFile)[-1]
    html += "\n"
    html += "    spec file: %s\n"                     % href
    html += "    <br />\n"
    html += "\n"
    if len(problem_scans) > 0:
        html += "    <h2>%d scan(s) with plotting problems</h2>\n" % len(problem_scans)
        html += "\n"
        html += "    <ul>\n"
        for item in problem_scans:
            html += "    <li>%s</li>\n" % item
        html += "    </ul>\n"
    html += "\n"
    html += "    <h2>%d plotted scan(s)</h2>\n" % len(plotted_scans)
    html += "\n"
    html += "\n".join(plotted_scans)
    html += "\n"
    html += "\n"
    html += "    <hr />\n"
    ref = 'page created: ' + str(datetime.datetime.now())
    ref += ', specplot_gallery documentation: <a href="%s">%s</a>' % (DOC_URL, DOC_URL)
    html += "    <center><small>%s</small></center>\n" % ref
    html += "\n"
    html += "  </body>\n"
    html += "</html>\n"
    return html


def logger(message):
    """
    log a message or report from this module

    :param str message: text to be logged
    """
    now = str(datetime.datetime.now())
    name = os.path.basename(sys.argv[0])
    pid = os.getpid()
    text = "(%d,%s,%s) %s" % (pid, name, now, message)
    logging.info(text)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def main():
    import argparse
    global SPECPLOTS_DIR

    doc = __doc__.strip().splitlines()[0]
    p = argparse.ArgumentParser(description=doc)
    
    p.add_argument(
        'paths',  
        nargs='+',  
        help="SPEC data file name(s) or directory(s) with SPEC data files")
    
    p.add_argument(
        '-r', 
        action='store_true', 
        default=False,
        dest='reverse_chronological',
        help='sort images from each data file in reverse chronolgical order')

    pwd = os.getcwd()
    msg = 'base directory for output'
    msg += ' (default:' + pwd + ')'
    p.add_argument('-d', '--dir', help=msg)

    args = p.parse_args()

    specplots_dir = args.dir or pwd
    if not os.path.exists(specplots_dir):
        raise DirectoryNotFoundError(specplots_dir)
    if not os.path.isdir(specplots_dir):
        raise PathIsNotDirectoryError(specplots_dir)
    
    file_list = []

    def only_accept_spec_files(fname):
        if os.path.exists(fname) and spec.is_spec_file(fname):
            file_list.append(fname)

    for item in args.paths:
        if os.path.exists(item):
            if os.path.isfile(item):
                only_accept_spec_files(item)
            elif os.path.isdir(item):
                for subitem in os.listdir(item):
                    only_accept_spec_files(os.path.join(item, subitem))

    log_file = os.path.join(specplots_dir, 'specplot_files_processing.log')
    logging.basicConfig(filename=log_file, level=logging.INFO)

    logger('>'*10 + ' starting')
    # TODO: do not start this process if it is running from previous call
    PlotSpecFileScans(
        file_list, 
        specplots_dir, 
        reverse_chronological=args.reverse_chronological)
    logger('<'*10 + ' finished')


def developer():
    """
    supply a file and a directory as command-line arguments to "paths"
    """
    import tempfile
    
    tempdir = tempfile.mkdtemp()
    sys.argv.append('-d')
    sys.argv.append(tempdir)
    sys.argv.append(os.path.join('data', '02_03_setup.dat'))
    sys.argv.append(os.path.join('..', '..', 'tests','data'))
    main()
    logging.disable(logging.CRITICAL)
    logging.shutdown()
    shutil.rmtree(tempdir)
    logging.disable(logging.NOTSET)


if __name__ == '__main__':
    main()
    # developer()



#!/usr/bin/env python

#-----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     prjemian@gmail.com
# :copyright: (c) 2014-2019, Pete R. Jemian
#
# Distributed under the terms of the Creative Commons Attribution 4.0 International Public License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

"""
Plot the data from scan N in a SPEC data file

.. autosummary::

    ~Selector
    ~ImageMaker
    ~LinePlotter
    ~MeshPlotter
    ~openSpecFile

Exceptions:

.. autosummary::

    ~NoDataToPlot
    ~NotPlottable
    ~ScanAborted
    ~UnexpectedObjectTypeError

"""

import os
import numpy

from . import charts
from . import spec             # read SPEC data files
from . import singletons
from . import utils


class UnexpectedObjectTypeError(RuntimeError): 
    'Exception: incorrect Python object type: programmer error'
    pass

class ScanAborted(RuntimeWarning): 
    'Exception: Scan aborted before all points acquired'
    pass

class NotPlottable(ValueError): 
    'Exception: No plottable data for this scan'
    pass

class NoDataToPlot(ValueError): 
    'Exception: scan aborted before any points gathered or data not present in SPEC file'
    pass

ABORTED_ATTRIBUTE_TEXT = '_aborted_'


class Selector(singletons.Singleton):
    """
    associate SPEC scan macro names with image makers
    
    :image maker: subclass of :class:`ImageMaker`
    
    To include a custom image maker from outside this module,
    create the subclass and then add it to an instance of this
    class.  Such as this plotter that defaults to a logarithmic 
    scale for the X axis for all `logxscan` macros:
    
        from spec2nexus import specplot
        
        class LogX_Plotter(specplot.ImageMaker):

            def x_log(self):
                return True
        
        # ...
        
        selector = specplot.Selector()
        selector.add('logxscan', LogX_Plotter)
        
        # ...
        
        image_maker = specplot.Selector().auto(scan)
        plotter = image_maker()
        plotter.plot_scan(scan, fullPlotFile)

    This class is a singleton which means you will always get the same
    instance when you call this class many times in your program.
    
    .. autosummary::
    
        ~auto
        ~add
        ~update
        ~get
        ~exists
        ~default

    """
    default_key = '__default__'
    
    def __init__(self):
        self.db = {}
        self.add(self.default_key, LinePlotter, default=True)
    
    def auto(self, scan):
        """
        automatically choose a scan image maker based on the SPEC scan macro
        
        Selection Rules:
        
        * macro ends with "scan": use :class:`LinePlotter`
        * macro ends with "mesh": use :class:`MeshPlotter`
        * default: use default image maker (initially :class:`LinePlotter`)
        """
        if not isinstance(scan, (spec.SpecDataFileScan, )):
            msg = 'expected a SPEC scan object, received: '
            msg += str(scan)
            raise UnexpectedObjectTypeError(msg)
        
        macro = scan.get_macro_name()
        if self.exists(macro):
            return self.get(macro)

        # adapt for different scan macros
        image_maker = self.default()
        if macro == 'hklscan':
            image_maker = HKLScanPlotter
        elif macro.lower().endswith('scan'):     
            image_maker = LinePlotter
        elif macro.lower().endswith('mesh'):
            image_maker = MeshPlotter

        # register this macro name
        self.add(macro, image_maker)
        return image_maker
    
    def add(self, key, value, default=False):
        """
        register a new value by key
        
        :param str key: name of key, typically the macro name
        :raises KeyError: if key exists
        :raises UnexpectedObjectTypeError: if value is not subclass of :class:`ImageMaker`
        """
        if self.exists(key):
            raise KeyError('key exists: ' + key)
        if not issubclass(value, ImageMaker):
            msg = 'expected subclass of ImageMaker, received type: '
            msg += type(value).__name__
            raise UnexpectedObjectTypeError(msg)

        self.db[key] = value
        if default:
            self.db[self.default_key] = value

        return value
    
    def update(self, key, value, default=False):
        """
        replace an existing key with a new value
        
        :param str key: name of key, typically the macro name
        :raises KeyError: if key does not exist
        :raises UnexpectedObjectTypeError: if value is not subclass of :class:`ImageMaker`
        """
        if not self.exists(key):
            raise KeyError('key does not exist: ' + key)
        if not issubclass(value, ImageMaker):
            msg = 'expected subclass of ImageMaker, received type: '
            msg += type(value).__name__
            raise UnexpectedObjectTypeError(msg)

        self.db[key] = value
        if default:
            self.db[self.default_key] = value

        return value
    
    def get(self, key):
        """
        return a value by key
        
        :returns: subclass of :class:`ImageMaker` or `None` if key not found
        """
        return self.db.get(key)
    
    def exists(self, key):
        """
        is the key known? 
        """
        return key in self.db
    
    def default(self):
        """
        retrieve the value of the default key 
        """
        return self.get(self.default_key)


class ImageMaker(object):
    r"""
    superclass to handle plotting of data from a SPEC scan
    
    .. rubric:: Internal data model

    :signal: name of the 'signal' data (default data to be plotted)
    :data: values of various collected arrays {label: array}
    :axes: names of the axes of signal data
    
    .. rubric:: USAGE:
    
    #. Create a subclass of :class:`ImageMaker`
    #. Override any of these methods:

       .. autosummary::
        
            ~data_file_name
            ~make_image
            ~plottable
            ~plot_options
            ~retrieve_plot_data
    
    .. rubric:: EXAMPLE
    
    ::

        class LinePlotter(ImageMaker):
            '''create a line plot'''
            
            def make_image(self, plotFile):
                '''
                make MatPlotLib chart image from the SPEC scan
                
                :param obj plotData: object returned from :meth:`retrieve_plot_data`
                :param str plotFile: name of image file to write
                '''
                assert(self.signal in self.data)
                assert(len(self.axes) == 1)
                assert(self.axes[0] in self.data)

                y = self.data[self.signal]
                x = self.data[self.axes[0]]
                xy_plot(x, y,  plotFile, 
                       title = self.plot_title(),
                       plot_subtitle = self.plot_subtitle(),
                       xtitle = self.x_title(),
                       ytitle = self.y_title(),
                       xlog = self.x_log(),
                       ylog = self.y_log(),
                       timestamp_str = self.timestamp())

        sfile = specplot.openSpecFile(specFile)
        scan = sfile.getScan(scan_number)
        plotter = LinePlotter()
        plotter.plot_scan(scan, plotFile, y_log=True)
    
    """

    def __init__(self):
        self.scan = None
        self.settings = self._initialize_settings_()
        self.signal = None
        self.axes = []
        self.data = {}
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support methods that a subclass might override

    def data_file_name(self):
        """
        the name of the file with the actual data
        
        Usually, this is the SPEC data file
        but it *could* be something else
        """
        return self.scan.header.parent.fileName  # self.scan.specFile

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan
        
        The data to be plotted are provided in:
        
        * `self.signal`
        * `self.axes`
        * `self.data`
        
        :param str plotFile: name of image file to write
        """
        raise NotImplementedError('must implement make_image() in each subclass')
    
    def plottable(self):
        """
        can this data be plotted as expected?
        """
        return False    # override in subclass with specific tests
    
    def plot_options(self):
        """
        re-define any plot options in a subclass 
        """
        pass
    
    def retrieve_plot_data(self):
        """
        retrieve default plottable data from spec data file and store locally
        
        This method must retrieve the data to be plotted, either from the
        SPEC data file scan or from a file which name is provided
        in the scan detalis.
        
        These attributes must be set by this method:

        :data: dictionary containing values of the various collected arrays {label: array}
        :signal: name of the 'signal' data (default data to be plotted)
        :axes: names of the axes of signal data
        
        .. rubric:: Example data
        
        ::
        
            self.data = {
                'angle': [1, 2, 3, 4, 5],
                'counts': [0. 2. 55. 3. 0]}
            self.signal = 'counts'
            self.axes = ['angle']
        
        Raise any of these exceptions as appropriate:

        .. autosummary::
        
            ~NoDataToPlot
            ~NotPlottable
            ~ScanAborted

        """
        raise NotImplementedError('must implement retrieve_plot_data() in each subclass')
         
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support methods that will not need to be defined in a subclass
    
    def data_is_newer_than_plot(self, plotFile):
        """only proceed if mtime of SPEC data file is newer than plotFile"""
        mtime_sdf = os.path.getmtime(self.data_file_name())
        if os.path.exists(plotFile):
            mtime_pf = os.path.getmtime(plotFile)
        else:
            mtime_pf = 0

        return mtime_sdf > mtime_pf

    def plot_scan(self, scan, plotFile, maker=None):
        """
        make an image plot of the data in the scan

        :param obj scan: instance of :class:`~spec2nexus.spec.SpecDataFileScan`
        :param str plotFile: file name for plot output
        """
        if not isinstance(scan, (spec.SpecDataFileScan, )):
            raise UnexpectedObjectTypeError('scan object not a SpecDataFileScan')
        if hasattr(scan, ABORTED_ATTRIBUTE_TEXT):
            match_text = 'Scan aborted after 0 points.'
            if scan.__getattribute__(ABORTED_ATTRIBUTE_TEXT) == match_text:
                raise ScanAborted(match_text)

        self.scan = scan

        self.set_plot_title(self.plot_title() or self.data_file_name())
        self.set_plot_subtitle(
            self.plot_subtitle() or 
            '#' + str(self.scan.scanNum) + ': ' + self.scan.scanCmd)
        self.set_timestamp(self.timestamp() or self.scan.date)

        try:
            self.retrieve_plot_data()
        except KeyError as _exc:
            if hasattr(self.scan, ABORTED_ATTRIBUTE_TEXT):
                raise ScanAborted(self.scan.__getattribute__(ABORTED_ATTRIBUTE_TEXT))
            raise _exc
        
        self.plot_options()
        
        if self.plottable() and self.data_is_newer_than_plot(plotFile):
            self.make_image(plotFile)
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # support the self.settings dictionary with get & set methods
    
    def _initialize_settings_(self):
        """
        initial values are set to `None`
        
        subclasses that set a value should first check if the value has already been set
        unless explicitly replacing any customizations by the user
        """
        return dict(
            title = None,
            subtitle = None,
            x_title = None,
            y_title = None,
            x_log = None,
            y_log = None,
            z_log = None,
            timestamp = None,
        )
    
    def plot_title(self):
        'return the plot title'
        return self.settings['title']
    
    def set_plot_title(self, text):
        'set the plot title'
        self.settings['title'] = text
    
    def plot_subtitle(self):
        'return the plot_subtitle'
        return self.settings['subtitle']
    
    def set_plot_subtitle(self, text):
        'set the plot_subtitle'
        self.settings['subtitle'] = text
    
    def x_title(self):
        'return the title for the X axis'
        return self.settings['x_title']
    
    def set_x_title(self, text):
        'set the x axis title'
        self.settings['x_title'] = text
    
    def y_title(self):
        'return the title for the Y axis'
        return self.settings['y_title']
    
    def set_y_title(self, text):
        'set the y axis title'
        self.settings['y_title'] = text
    
    def x_log(self):
        'boolean: should the X axis be plotted on a log scale?'
        return self.settings['x_log']
    
    def set_x_log(self, choice):
        'set the x axis logarithmic if True'
        self.settings['x_log'] = choice
    
    def y_log(self):
        'boolean: should the Y axis be plotted on a log scale?'
        return self.settings['y_log']
    
    def set_y_log(self, choice):
        'set the y axis logarithmic if True'
        self.settings['y_log'] = choice
    
    def z_log(self):
        'boolean: should the Z axis (image) be plotted on a log scale?'
        return self.settings['z_log']
    
    def set_z_log(self, choice):
        'set the z axis (image) logarithmic if True'
        self.settings['z_log'] = choice
    
    def timestamp(self):
        'return the time of this scan as a string'
        return self.settings['timestamp']
    
    def set_timestamp(self, text):
        'set the plot time stamp'
        self.settings['timestamp'] = text


class LinePlotter(ImageMaker):
    """
    create a line plot
    """

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan
        
        :param str plotFile: name of image file to write
        """
        assert(self.signal in self.data)
        assert(len(self.axes) == 1)
        assert(self.axes[0] in self.data)

        y = self.data[self.signal]
        x = self.data[self.axes[0]]
        ts = self.timestamp()

        charts.xy_plot(
            x, 
            y,  
            plotFile, 
            title = self.plot_title(),
            subtitle = self.plot_subtitle(),
            xtitle = self.x_title(),
            ytitle = self.y_title(),
            xlog = self.x_log(),
            ylog = self.y_log(),
            timestamp_str = ts)

    def plottable(self):
        """
        can this data be plotted as expected?
        """
        if self.signal in self.data:
            signal = self.data[self.signal]
            if signal is not None and len(signal) > 0 and len(self.axes) == 1:
                if len(signal) == len(self.data[self.axes[0]]):
                    return True
        return False
    
    def plot_options(self):
        """
        define the settings for this, accepting any non-default values first
        """
        self.x_title() or self.set_x_title(self.axes[0])
        self.y_title() or self.set_y_title(self.signal)
        self.x_log() or self.set_x_log(False)
        self.y_log() or self.set_y_log(False)
        self.set_z_log(False)
    
    def retrieve_plot_data(self):
        """retrieve default data from spec data file"""
        # plot last column v. first column
        assert(isinstance(self.scan, spec.SpecDataFileScan))
        self.signal = self.scan.column_last
        if self.signal not in self.scan.data:
            raise NoDataToPlot(str(self.scan))
        self.axes = [self.scan.column_first,]
        self.data = {label: self.scan.data.get(label) for label in self.scan.L if label in self.scan.data}


class HKLScanPlotter(LinePlotter):
    """
    create a line plot from hklscan macros
    """

    def retrieve_plot_data(self):
        """retrieve default data from spec data file"""
        # standard hklscan macro handling
        # find the real scan axis, the one that changes
        for axis in 'H K L'.split():
            data = self.scan.data.get(axis)
            if data is None:
                continue
            # could compare start & end from scanCmd, this looks simpler
            if min(data) != max(data):
                # tell it to use this axis instead
                self.axes = [axis,]
                break

        # if not found, default changes nothing
        if data is None:
            raise NotPlottable('no data in scan: ' + str(self.scan))

        if len(self.axes) == 0:
            # issue #99:  file: lmn40.spe, scan 244, hkl all fixed
            axis = 'data point number'
            data = range(1, 1+len(self.scan.data['H']))
            self.scan.data[axis] = data
            self.set_x_title(axis + ' (hkl all held constant)')
            self.axes = [axis,]

        self.scan.column_first = axis
        self.data[axis] = data
        self.signal = self.scan.column_last
        self.data[self.signal] = self.scan.data[self.signal]


class MeshPlotter(ImageMaker):
    """
    create a mesh plot (2-D image)

    ..rubric:: References:
    
    :mesh 2-D parser: http://www.certif.com/spec_help/mesh.html
        
        ::
        
            mesh motor1 start1 end1 intervals1 motor2 start2 end2 intervals2 time
    
    :hklmesh 2-D parser: http://www.certif.com/spec_help/hklmesh.html
        
        ::
        
            hklmesh Q1 start1 end1 intervals1 Q2 start2 end2 intervals2 time 
               
    """
    # see code in: writer.Writer.mesh()        self._mesh_(scan)

    def make_image(self, plotFile):
        """
        make MatPlotLib chart image from the SPEC scan
        
        :param str plotFile: name of image file to write
        """
        if len(self.axes) == 2:
            image = self.data[self.signal]
            self.set_plot_subtitle(
                '%s,  %s' % (self.signal, 
                             self.scan.raw.splitlines()[0])
                )
            self.set_x_title(self.axes[0])
            self.set_y_title(self.axes[1])
            
            charts.make_png(
                image, 
                plotFile,
                [self.data[axis] for axis in self.axes],
                title = self.plot_title(),
                subtitle = self.plot_subtitle(),
                timestamp_str = self.timestamp(),
                xtitle = self.x_title(),
                ytitle = self.y_title(),
                log_image = self.z_log(),
                )
        elif len(self.axes) == 1:
            # fallback to 1-D plot
            y = self.data[self.signal]
            x = self.data[self.axes[0]]
            charts.xy_plot(
                x, 
                y,  
                plotFile, 
                title = self.plot_title(),
                subtitle = self.plot_subtitle(),
                xtitle = self.x_title(),
                ytitle = self.y_title(),
                xlog = self.x_log(),
                ylog = self.y_log(),
                timestamp_str = self.timestamp())
    
    def plottable(self):
        """
        can this data be plotted as expected?
        """
        try:
            assert(self.signal in self.data)
            signal = numpy.array(self.data[self.signal])
            assert(len(self.axes) in (0,len(signal.shape)))
            for order, axis in enumerate(reversed(self.axes)):
                assert(axis in self.data)
                assert(signal.shape[order] == len(self.data[axis]))
        except Exception:
            return False
        return True
    
    def plot_options(self):
        """
        define the settings for this, accepting any non-default values first
        """
        if len(self.axes) == 1:
            self.x_title() or self.set_x_title(self.axes[0])
            self.y_title() or self.set_y_title(self.signal)
        elif len(self.axes) == 2:
            self.x_title() or self.set_x_title(self.axes[1])
            self.y_title() or self.set_y_title(self.axes[0])
        self.x_log() or self.set_x_log(False)
        self.y_log() or self.set_y_log(False)
        self.z_log() or self.set_z_log(False)

    def retrieve_plot_data(self):
        """retrieve default data from spec data file
        
        data parser for 2-D mesh and hklmesh
        """
        label1, start1, end1, intervals1, label2, start2, end2, intervals2, time = self.scan.scanCmd.split()[1:]
        if label1 not in self.scan.data:
            label1 = self.scan.L[0]      # mnemonic v. name
        if label2 not in self.scan.data:
            label2 = self.scan.L[1]      # mnemonic v. name
        axis1 = self.scan.data.get(label1)
        axis2 = self.scan.data.get(label2)
        intervals1, intervals2 = map(int, (intervals1, intervals2))
        # unused: start1, end1, start2, end2, time = map(float, (start1, end1, start2, end2, time))

        if len(axis1) < intervals1 and min(axis2) == max(axis2):
            # stopped scan before second row started, 1-D plot is better (issue #82)
            self.axes = [label1,]
            self.signal = self.scan.column_last
            self.data[label1] = self.scan.data[label1]
            self.data[self.signal] = self.scan.data[self.signal]
            return

        axis1 = axis1[0:intervals1+1]
        self.data[label1] = axis1    # 1-D array

        axis2 = [axis2[row] for row in range(len(axis2)) if row % (intervals1+1) == 0]
        self.data[label2] = axis2    # 1-D array

        column_labels = self.scan.L
        column_labels.remove(label1)    # special handling
        column_labels.remove(label2)    # special handling
        if self.scan.scanCmd.startswith('hkl'):
            # find the reciprocal space axis held constant
            label3 = [key for key in ('H', 'K', 'L') if key in column_labels][0]
            self.data[label3] = self.scan.data.get(label3)[0]    # constant

        # build 2-D data objects (do not build label1, label2, [or label3] as 2-D objects)
        data_shape = [len(axis2), len(axis1)]
        for label in column_labels:
            if label not in self.data:
                axis = numpy.array( self.scan.data.get(label) )
                self.data[label] = utils.reshape_data(axis, data_shape)
            else:
                pass

        self.signal = utils.clean_name(self.scan.column_last)
        self.axes = [label1, label2]
    
        if spec.MCA_DATA_KEY in self.scan.data:    # 3-D array(s)
            # save each spectrum
            for key, spectrum in sorted(self.scan.data[spec.MCA_DATA_KEY].items()):
                num_channels = len(spectrum[0])
                data_shape.append(num_channels)
                mca = numpy.array(spectrum)
                data = utils.reshape_data(mca, data_shape)
                channels = range(1, num_channels+1)
                ds_name = '_' + key + '_'
                self.data[ds_name] = data
                self.data[ds_name+'channel_'] = channels


# class NeXusPlotter(ImageMaker):    # TODO: issue #92
#     """
#     create a plot from a NeXus HDF5 data file
#     """
#     
#     def retrieve_plot_data(self):
#         """retrieve default data from spec data file"""
#         raise NotImplementedError(self.__class__.__name__ + '() is not ready')
# 
#     def make_image(self, plotFile):
#         """
#         make image file from the SPEC scan
#         
#         :param str plotFile: name of image file to write
#         """
#         raise NotImplementedError(self.__class__.__name__ + '() is not ready')


def openSpecFile(specFile):
    """
    convenience routine so that others do not have to `import spec2nexus.spec`
    """
    sd = spec.SpecDataFile(specFile)
    return sd


def main():
    import argparse
    doc = __doc__.strip().splitlines()[0]
    p = argparse.ArgumentParser(description=doc)
    p.add_argument('specFile',    help="SPEC data file name")
    p.add_argument('scan_number', help="scan number in SPEC file", type=str)
    p.add_argument('plotFile',    help="output plot file name")
    args = p.parse_args()
    
    sfile = openSpecFile(args.specFile)
    scan = sfile.getScan(args.scan_number)
    image_maker = Selector().auto(scan)
    plotter = image_maker()
    plotter.plot_scan(scan, args.plotFile)


if __name__ == '__main__':
    main()
